Package: SpectriPy
Title: Enhancing Cross-Language Mass Spectrometry Data Analysis with R
        and Python
Version: 1.0.0
Description: The SpectriPy package allows integration of Python-based MS
    analysis code with the Spectra package. Spectra objects can be converted
    into Python MS data structures. In addition, SpectriPy integrates
    and wraps the similarity scoring and processing/filtering functions
    from the Python matchms package into R.
Authors@R: c(person(given = "Michael", family = "Witting",
                    role = c("aut"),
                    email = "michael.witting@helmholtz-muenchen.de",
                    comment = c(ORCID = "0000-0002-1462-4426")),
             person(given = "Johannes", family = "Rainer",
                    email = "Johannes.Rainer@eurac.edu",
                    role = c("aut", "cre"),
                    comment = c(ORCID = "0000-0002-6977-7147")),
             person(given = "Carolin", family = "Huber",
                    email = "Carolin-elisabeth.huber@ufz.de",
                    role = c("aut"),
                    comment = c(ORCID = "0000-0002-9355-8948")),
             person(given = "Helge", family = "Hecht",
                    email = "helge.hecht@recetox.muni.cz",
                    role = c("ctb"),
                    comment = c(ORCID = "0000-0001-6744-996X")),
             person(given = "Marilyn", family = "De Graeve",
                    email = "marilyn.degraeve@eurac.edu",
                    role = c("aut"),
                    comment = c(ORCID = "0000-0001-6916-401X")),
             person(given = "Wout", family = "Bittremieux",
                    email = "wout.bittremieux@uantwerpen.be",
                    role = c("aut"),
                    comment = c(ORCID = "0000-0002-3105-1359")),
             person(given = "Thomas", family = "Naake",
                    email = "thomasnaake@googlemail.com",
                    role = c("aut"),
                    comment = c(ORCID = "0000-0001-7917-5580")),
             person(given = "Victor", family = "Chrone",
                    email = "vchrone@bmb.sdu.dk",
                    role = c("ctb"),
                    comment = c(ORCID = "0009-0007-2121-4066")),
             person(given = "Matthias", family = "Anagho-Mattanovich",
                    email = "matthias.mattanovich@sund.ku.dk",
                    role = c("ctb"),
                    comment = c(ORCID = "0000-0001-7561-7898")),
             person(given = "Pierre", family = "Marchal",
                    email = "pierre.marchal@unibe.ch",
                    role = c("ctb"),
                    comment = c(ORCID = "0009-0006-6567-6257")),
             person(given = "Philippine", family = "Louail",
                    email = "philippine.louail@eurac.edu",
                    role = c("ctb"),
                    comment = c(ORCID = "0009-0007-5429-6846"))
             )
Depends: R (>= 4.4.0), reticulate (>= 1.42.0)
Imports: Spectra (>= 1.19.9), IRanges, S4Vectors, MsCoreUtils,
        ProtGenerics, methods
Suggests: testthat, quarto, MsBackendMgf, MsDataHub, mzR, knitr,
        BiocStyle
License: Artistic-2.0
BugReports: https://github.com/RforMassSpectrometry/SpectriPy/issues
URL: https://github.com/RforMassSpectrometry/SpectriPy
biocViews: Infrastructure, Metabolomics, MassSpectrometry, Proteomics
Encoding: UTF-8
SystemRequirements: python (>= 3.12), pandoc, quarto
VignetteBuilder: quarto
Roxygen: list(markdown = TRUE)
RoxygenNote: 7.3.3
Collate: 'conversion.R' 'MsBackendPython.R' 'compareSpectriPy.R'
        'filterSpectriPy.R' 'zzz.R'
git_url: https://git.bioconductor.org/packages/SpectriPy
git_branch: RELEASE_3_22
git_last_commit: 017658a
git_last_commit_date: 2025-10-29
Repository: Bioconductor 3.22
Date/Publication: 2025-10-29
NeedsCompilation: no
Packaged: 2025-10-30 06:50:09 UTC; biocbuild
Author: Michael Witting [aut] (ORCID: <https://orcid.org/0000-0002-1462-4426>),
  Johannes Rainer [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-6977-7147>),
  Carolin Huber [aut] (ORCID: <https://orcid.org/0000-0002-9355-8948>),
  Helge Hecht [ctb] (ORCID: <https://orcid.org/0000-0001-6744-996X>),
  Marilyn De Graeve [aut] (ORCID:
    <https://orcid.org/0000-0001-6916-401X>),
  Wout Bittremieux [aut] (ORCID: <https://orcid.org/0000-0002-3105-1359>),
  Thomas Naake [aut] (ORCID: <https://orcid.org/0000-0001-7917-5580>),
  Victor Chrone [ctb] (ORCID: <https://orcid.org/0009-0007-2121-4066>),
  Matthias Anagho-Mattanovich [ctb] (ORCID:
    <https://orcid.org/0000-0001-7561-7898>),
  Pierre Marchal [ctb] (ORCID: <https://orcid.org/0009-0006-6567-6257>),
  Philippine Louail [ctb] (ORCID:
    <https://orcid.org/0009-0007-5429-6846>)
Maintainer: Johannes Rainer <Johannes.Rainer@eurac.edu>
