# Generated by roxygen2: do not edit by hand

S3method(r_to_py,Spectra)
export(CosineGreedy)
export(CosineHungarian)
export(ModifiedCosine)
export(MsBackendPy)
export(NeutralLossesCosine)
export(compareSpectriPy)
export(defaultSpectraVariableMapping)
export(filterSpectriPy)
export(normalize_intensities)
export(pyspec_to_rspec)
export(reindex)
export(remove_peaks_around_precursor_mz)
export(rspec_to_pyspec)
export(select_by_intensity)
export(select_by_mz)
export(setSpectraVariableMapping)
exportMethods("spectraData<-")
exportMethods("spectraVariableMapping<-")
exportMethods(compareSpectriPy)
exportMethods(filterSpectriPy)
exportMethods(spectraVariableMapping)
importFrom(IRanges,NumericList)
importFrom(MsCoreUtils,rbindFill)
importFrom(MsCoreUtils,vapply1l)
importFrom(S4Vectors,DataFrame)
importFrom(Spectra,MsBackendMemory)
importFrom(Spectra,coreSpectraVariables)
importFrom(Spectra,fillCoreSpectraVariables)
importFrom(Spectra,spectrapply)
importFrom(methods,"slot<-")
importFrom(methods,as)
importFrom(methods,is)
importFrom(methods,new)
importFrom(reticulate,import)
importFrom(reticulate,iterate)
importFrom(reticulate,np_array)
importFrom(reticulate,py)
importFrom(reticulate,py_available)
importFrom(reticulate,py_config)
importFrom(reticulate,py_del_attr)
importFrom(reticulate,py_dict)
importFrom(reticulate,py_eval)
importFrom(reticulate,py_get_attr)
importFrom(reticulate,py_has_attr)
importFrom(reticulate,py_require)
importFrom(reticulate,py_run_string)
importFrom(reticulate,py_set_attr)
importFrom(reticulate,py_to_r)
importFrom(reticulate,r_to_py)
importMethodsFrom(ProtGenerics,"intensity<-")
importMethodsFrom(ProtGenerics,"mz<-")
importMethodsFrom(ProtGenerics,"peaksData<-")
importMethodsFrom(ProtGenerics,"spectraData<-")
importMethodsFrom(ProtGenerics,acquisitionNum)
importMethodsFrom(ProtGenerics,backendInitialize)
importMethodsFrom(ProtGenerics,centroided)
importMethodsFrom(ProtGenerics,collisionEnergy)
importMethodsFrom(ProtGenerics,dataOrigin)
importMethodsFrom(ProtGenerics,extractByIndex)
importMethodsFrom(ProtGenerics,intensity)
importMethodsFrom(ProtGenerics,isolationWindowLowerMz)
importMethodsFrom(ProtGenerics,isolationWindowTargetMz)
importMethodsFrom(ProtGenerics,isolationWindowUpperMz)
importMethodsFrom(ProtGenerics,msLevel)
importMethodsFrom(ProtGenerics,mz)
importMethodsFrom(ProtGenerics,peaksData)
importMethodsFrom(ProtGenerics,polarity)
importMethodsFrom(ProtGenerics,precScanNum)
importMethodsFrom(ProtGenerics,precursorCharge)
importMethodsFrom(ProtGenerics,precursorIntensity)
importMethodsFrom(ProtGenerics,precursorMz)
importMethodsFrom(ProtGenerics,rtime)
importMethodsFrom(ProtGenerics,scanIndex)
importMethodsFrom(ProtGenerics,smoothed)
importMethodsFrom(ProtGenerics,spectraData)
importMethodsFrom(ProtGenerics,spectraNames)
importMethodsFrom(ProtGenerics,spectraVariables)
importMethodsFrom(ProtGenerics,tic)
importMethodsFrom(S4Vectors,isEmpty)
importMethodsFrom(Spectra,"spectraVariableMapping<-")
importMethodsFrom(Spectra,Spectra)
importMethodsFrom(Spectra,intensity)
importMethodsFrom(Spectra,mz)
importMethodsFrom(Spectra,peaksData)
importMethodsFrom(Spectra,peaksVariables)
importMethodsFrom(Spectra,spectraData)
importMethodsFrom(Spectra,spectraVariableMapping)
importMethodsFrom(Spectra,spectrapply)
importMethodsFrom(methods,show)
